/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2012 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.

*/

#ifndef quantlib_buffered_data_factory_hpp 
#define quantlib_buffered_data_factory_hpp 

#include <ql/errors.hpp>

#include <boost/thread.hpp>
#include <boost/function.hpp>
#include <boost/lockfree/ringbuffer.hpp>
#include <boost/smart_ptr/detail/atomic_count.hpp>

#include <stdio.h>

namespace QuantLib {

    template <class ResultType>
    class BufferedDataFactory {
      public:
        BufferedDataFactory(const boost::function<ResultType ()>& f,
                            size_t bufferSize = 1000u) 
        : n_(0), f_(f),
          buffer_(bufferSize) {

            QL_REQUIRE(bufferSize>1u, "buffer size must be greater than one");

            thread_ = boost::shared_ptr<boost::thread>(
                new boost::thread(boost::ref(*this)));            
        }
        
        ~BufferedDataFactory() {
            ++n_;

            cond_.notify_one();
            thread_->join();
        }
        
        ResultType next() {
            ResultType retVal;

            while (!buffer_.dequeue(retVal)) {
                boost::lock_guard<boost::mutex> guard(mutex_);
                cond_.notify_one();
            };

            return retVal;
        }

        void operator()() {
            boost::unique_lock<boost::mutex> lock(mutex_);

            while (!((long) n_)) {
                result_ = f_();

                while (!buffer_.enqueue(result_) && !((long) n_)) {
                    cond_.wait(lock);
                }
            }
        }

     private:
        ResultType result_;
        boost::detail::atomic_count n_;
        const boost::function<ResultType ()> f_;

        boost::mutex mutex_;
        boost::condition_variable cond_;
        boost::shared_ptr<boost::thread> thread_;
        boost::lockfree::ringbuffer<ResultType, 0u> buffer_;
    };
}

#endif
